from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Table, Column, Boolean, Integer, Text
from sqlalchemy import String, DateTime, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy import CheckConstraint, UniqueConstraint
from sqlalchemy.orm import object_session
import datetime

Base = declarative_base()

class Section(Base):
    __tablename__ = 'section'
    __table_args__ = (
        UniqueConstraint('parent_id', 'number'),
        CheckConstraint('(parent_id IS NULL AND id = 0) OR (parent_id IS NOT NULL AND id <> 0)'),
        CheckConstraint('parent_id <> id')
    )
    
    __export_headers__ = ['number', 'name', 'abbreviation',]
    
    id = Column(Integer, primary_key=True, index=True)
    number = Column(Integer, CheckConstraint('number >= 0'), nullable=False, index=True)
    name = Column(String, index=True, nullable=False)
    abbreviation = Column(String, default='', nullable=False)
    parent_id = Column(Integer, ForeignKey('section.id'))
    
    parent = relationship("Section", remote_side=[id])
    subsections = relationship("Section", cascade='all, delete-orphan')
    books = relationship('Book', back_populates="section")
    
    @property
    def unpadded_full_number(self):
        number = []
        section = self
        while section.parent is not None:
            number.append(str(section.number))
            section = section.parent
        return '-'.join(reversed(number))
    
    @property
    def full_number(self):
        cls = type(self)
        digits = len(str(object_session(self).query(cls).order_by(
            cls.number.desc()).first().number))
        template = '{:0{digits}}'
        number = []
        section = self
        while section.parent is not None:
            number.append(template.format(section.number, digits=digits))
            section = section.parent
        return '-'.join(reversed(number))
    
    @property
    def full_abbreviation(self):
        abbr = []
        section = self
        while section.parent is not None:
            abbr.append(str(section.abbreviation))
            section = section.parent
        return '-'.join(reversed(abbr))
    
    def __repr__(self):
        return "<Section {0.unpadded_full_number} {0.name} {0.abbreviation}>".format(self)
    
    def __iter__(self):
        return [self.unpadded_full_number, self.name, self.abbreviation].__iter__()
    
    @classmethod
    def from_values(cls, session, full_number, name, abbr, *args):
        numbers = full_number.split('-')
        parent = session.query(cls).get(0)
        for c in numbers[0:-1]:
            parent = session.query(cls).filter(cls.number == c,
                cls.parent == parent).one()
        
        return cls(number=numbers[-1], name=name, abbreviation=abbr,
            parent=parent)

book_author = Table('book_author', Base.metadata,
    Column('id', Integer, primary_key=True, index=True),
    Column('book_id', Integer, ForeignKey('book.id', ondelete='cascade'), nullable=False),
    Column('author_id', Integer, ForeignKey('author.id', ondelete='cascade'), nullable=False),
    UniqueConstraint('book_id', 'author_id')
)

class BookAuthor(Base):
    __table__ = book_author
    __mapper_args__ = {
            'confirm_deleted_rows': False
        }
    id = book_author.c.id
    book_id = book_author.c.book_id
    author_id = book_author.c.author_id
    
    book = relationship("Book")
    author = relationship("Author")

class Tag(Base):
    __tablename__ = 'tag'
    __table_args__ = (
            UniqueConstraint('text', 'book_id'),
        )
    
    __export_headers__ = ['text',]
    
    id = Column(Integer, primary_key=True, index=True)
    text = Column(String, index=True, nullable=False)
    book_id = Column(Integer, ForeignKey('book.id'), nullable=False)
    
    book = relationship("Book", back_populates="tags")
    
    def __str__(self):
        return self.text
    
    def __iter__(self):
        return [self.text].__iter__()
    
    @classmethod
    def from_values(cls, session, text, *args):
        return cls(text=text)

class Author(Base):
    __tablename__ = 'author'
    
    __export_headers__ = ['name', 'about',]
    
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, CheckConstraint("name <> ''"), unique=True, index=True, nullable=False)
    about = Column(Text, nullable=False, default='')
    
    books = relationship("Book", secondary=book_author, back_populates="authors")
    
    def __str__(self):
        return self.name
    
    def __iter__(self):
        return [self.name, self.about].__iter__()
    
    @classmethod
    def from_values(cls, session, name, about='', *args):
        return cls(name=name, about=about)

class Publisher(Base):
    __tablename__ = 'publisher'
    
    __export_headers__ = ['name', 'about',]
    
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, CheckConstraint("name <> ''"), unique=True, index=True, nullable=False)
    about = Column(Text, nullable=False, default='')
    
    books = relationship("Book", back_populates="publisher")
    
    def __str__(self):
        return self.name
    
    def __iter__(self):
        return [self.name, self.about].__iter__()
    
    @classmethod
    def from_values(cls, session, name, about='', *args):
        return cls(name=name, about=about)

class Library(Base):
    __tablename__ = 'library'
    
    __export_headers__ = ['name',]
    
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, CheckConstraint("name <> ''"), unique=True, index=True, nullable=False)
    
    books = relationship("Book", back_populates="library")
    
    def __str__(self):
        return self.name
    
    def __iter__(self):
        return [self.name].__iter__()
    
    @classmethod
    def from_values(cls, session, name, *args):
        return cls(name=name)

class Book(Base):
    __tablename__ = 'book'
    
    __export_headers__ = [
        'number', 'title', 'authors', 'publisher', 'publication year',
        'library', 'is electronic?', 'language', 'tags',
    ]
    
    id = Column(Integer, primary_key=True, index=True)
    section_id = Column(
            Integer,
            ForeignKey('section.id'),
            CheckConstraint('section_id <> 0'),
            nullable=False
        )
    number = Column(
            Integer,
            CheckConstraint('number > 0'),
            nullable=False,
            index=True
        )
    title = Column(
            String,
            CheckConstraint("title <> ''"),
            index=True,
            nullable=False
        )
    publication_year = Column(
            Integer,
            CheckConstraint('number > 0'),
            index=True
        )
    
    publisher_id = Column(Integer, ForeignKey('publisher.id'))
    library_id = Column(Integer, ForeignKey('library.id'))
    is_electronic = Column(Boolean,  nullable=False, index=True)
    language = Column(String, index=True, nullable=False)
    
    section = relationship("Section", back_populates="books")
    authors = relationship("Author", secondary=book_author, back_populates="books")
    publisher = relationship("Publisher", back_populates="books")
    library = relationship("Library", back_populates="books")
    borrows = relationship("Borrow", cascade='all, delete-orphan')
    tags = relationship("Tag", back_populates="book",
        cascade='all, delete-orphan')
    
    @property
    def unpadded_full_number(self):
        return '{}-{}'.format(self.section.unpadded_full_number, self.number)
    
    @property
    def full_number(self):
        cls = type(self)
        digits = len(str(object_session(self).query(cls).order_by(
            cls.number.desc()).first().number))
        return '{}-{:0{digits}}'.format(self.section.full_number, self.number,
            digits=digits)
    
    @property
    def authors_names(self):
        return '|'.join([c.name for c in self.authors])
    
    @property
    def tags_text(self):
        return ' '.join([c.text for c in self.tags])
    
    @property
    def publisher_name(self):
        return self.publisher.name if self.publisher is not None else ''
    
    @property
    def publication_year_str(self):
        return self.publication_year if self.publication_year is not None else ''
    
    @property
    def library_name(self):
        return self.library.name if self.library is not None else ''
    
    @property
    def copies(self):
        session = object_session(self)
        q = session.query(type(self)).filter(
            section_id=self.section_id, number=self.number)
        return q.all()
    
    @property
    def copies_count(self):
        session = object_session(self)
        q = session.query(type(self)).filter(
            section_id=self.section_id, number=self.number)
        return q.count()
    
    def __repr__(self):
        return "<Book {0.unpadded_full_number} {0.title}>".format(self)
    
    def __iter__(self):
        if self.publisher is not None:
            publisher = self.publisher.name
        else:
            publisher = ''
        
        if self.library is not None:
            library = self.library.name
        else:
            library = ''
        
        return [self.unpadded_full_number, self.title, self.authors_names,
            publisher, self.publication_year, library, int(self.is_electronic),
            self.language, self.tags_text].__iter__()
    
    @classmethod
    def from_values(cls, session, full_number, title, authors_names='',
            publisher='', publication_year='', library='', is_electronic=False,
            language='', tags_text='', *args):
        numbers = full_number.split('-')
        section = session.query(Section).get(0)
        for c in numbers[0:-1]:
            section = session.query(Section).filter(Section.number == c,
                Section.parent == section).one()
        
        authors = [c.strip() for c in authors_names.split('|')]
        authors = [c for c in authors if c]
        for idx, c in enumerate(authors):
            try:
                authors[idx] = session.query(Author).filter(
                    Author.name == c).one()
            except NoResultFound:
                authors[idx] = Author(name=c)
        
        if publisher:
            try:
                publisher = session.query(Publisher).filter(
                    Publisher.name == publisher).one()
            except NoResultFound:
                publisher = Publisher(name=publisher)
        else:
            publisher = None
        
        if not publication_year:
            publication_year = None
        
        if library:
            try:
                library = session.query(Library).filter(
                    Library.name == library).one()
            except NoResultFound:
                library = Library(name=library)
        else:
            library = None
        
        tags = [c.strip() for c in tags_text.split()]
        tags = [Tag(text=c) for c in tags if c]
        
        return cls(number=numbers[-1], section=section, title=title,
            authors=authors, publisher=publisher,
            publication_year=publication_year, library=library,
            is_electronic=int(is_electronic), language=language, tags=tags)

class Borrow(Base):
    __tablename__ = 'borrow'
    
    __export_headers__ = [
        'book number', 'borrower', 'contact', 'date', 'return date',
    ]
    
    id = Column(Integer, primary_key=True, index=True)
    book_id = Column(Integer, ForeignKey('book.id'), nullable=False)
    borrower = Column(String, CheckConstraint("borrower <> ''"), index=True, nullable=False)
    contact = Column(String, CheckConstraint("contact not like '%[^0-9]%' AND contact <> ''"), index=True, nullable=False)
    date = Column(DateTime, index=True, nullable=False, default=datetime.datetime.now())
    return_date = Column(DateTime, index=True)
    
    book = relationship("Book", back_populates='borrows')
    
    @property
    def date_str(self):
        return '{:%Y-%m-%d}'.format(self.date)
    
    @property
    def return_date_str(self):
        if self.return_date is not None:
            return '{:%Y-%m-%d}'.format(self.return_date)
        else:
            return ''
    
    def return_(self):
        self.return_date = datetime.datetime.now()
    
    def __repr__(self):
        return "<Borrow {0.book.unpadded_full_number} {0.borrower} {0.date:%Y%m%d}>".format(self)
    
    def __iter__(self):
        if self.return_date is not None:
            return_date = '{:%Y-%m-%d}'.format(self.return_date)
        else:
            return_date = ''
        return [self.book.unpadded_full_number, self.borrower, self.contact,
            '{:%Y-%m-%d}'.format(self.date), return_date].__iter__()
    
    @classmethod
    def from_values(cls, session, full_number, borrower, contact, date,
            return_date, *args):
        numbers = full_number.split('-')
        section = session.query(Section).get(0)
        for c in numbers[0:-1]:
            section = session.query(Section).filter(Section.number == c,
                Section.parent == section).one()
        book = session.query(Book).filter(Book.number == numbers[-1],
                Book.section == section).one()
        
        date = datetime.datetime.strptime(date, '%Y-%m-%d')
        
        if return_date:
            return_date = datetime.datetime.strptime(return_date, '%Y-%m-%d')
        else:
            return_date = None
        
        return cls(book=book, borrower=borrower, contact=contact, date=date,
            return_date=return_date)

def default_args(args, defaults, headers=[]):
    args = list(args)
    for c in range(len(defaults)):
        if len(args) < c:
            if defaults[c] is None:
                raise ValueError('`{}` missing'.format(headers[c]))
            else:
                args.append(defaults[c])
        elif args[c] == '':
            if defaults[c] is None:
                raise ValueError('`{}` missing'.format(headers[c]))
            else:
                args[c] = defaults[c]
    
    return args

CLASSES = [
        Library,
        Author,
        Publisher,
        Section,
        Book,
        Borrow,
    ]



from . import db
from .iqra_plugins import IqraCommand
from . import IqraCmd
from sqlalchemy import String
from sqlalchemy import func

class new(IqraCommand):
    'create a new section or book'
    
    class section(IqraCommand):
        'create a new section'
        
        args = (
            (['number'], dict(type=str, help='section number including all parent sections')),
            (['name'], dict(type=str, help='section name')),
            (['abbr'], dict(type=str, help='section abbreviation')),
        )
        
        def call(self, number, name, abbr):
            number = [int(c) for c in number.split('-')]
            
            session = self.util.get_scoped_session()
            
            section = self.util.get_root_section(session)
            for c in number[0:-1]:
                section = session.query(db.Section).filter(
                    db.Section.number == c, db.Section.parent == section).one()
            
            section = db.Section(number=number[-1],
                name=name, abbreviation=abbr, parent=section)
            
            session.add(section)
            return session
    
    class book(IqraCommand):
        'create a new book'
        
        args = (
            (['number'], dict(type=str, help='book number including all parent sections')),
            (['title'], dict(type=str, help='book title')),
            (['-p', '--publisher'], dict(type=str, help='publisher')),
            (['-a', '--authors'], dict(type=str, help='comma separated author names')),
            (['-l', '--library'], dict(type=str, help='library name')),
            (['-d', '--date'], dict(type=int, help='publication year')),
            (['-e', '--electronic'], dict(action='store_true', default=False, help='the book is an electronic book. opposite -ne')),
            (['-ne', '--not-electronic'], dict(action='store_false', dest='electronic', help='the book is a paper book. opposite -e')),
            (['--language'], dict(type=str, default='', help='language of the book')),
            (['--tags'], dict(type=str, default='', help='space seperated book tags')),
        )
        
        def call(self, number, title, publisher=None, authors=None,
                library=None, date=None, electronic=False, language='',
                tags=''):
            number = [int(c) for c in number.split('-')]
            
            session = self.util.get_scoped_session()
            section = self.util.get_root_section(session)
            for c in number[0:-1]:
                section = session.query(db.Section).filter(
                    db.Section.number == c, db.Section.parent == section).one()
            
            number = number[-1]
            
            if publisher is not None:
                obj = session.query(db.Publisher).filter(
                    db.Publisher.name == publisher).first()
                
                if obj is None:
                    obj = Publisher(name=publisher)
                
                publisher = obj
            
            if authors is not None:
                authors = [c.strip() for c in authors.split(',')]
                authors = [c for c in authors if c]
                for idx, c in enumerate(authors):
                    obj = session.query(db.Author).filter(
                        db.Author.name == c).first()
                    if obj is None:
                        obj = db.Author(name=c)
                    
                    authors[idx] = obj
            else:
                authors = []
            
            if library is not None:
                obj = session.query(db.Library).filter(
                    db.Library.name == library).first()
                
                if obj is None:
                    obj = Library(name=library)
                
                library = obj
            
            book = db.Book(number=number, section=section, title=title,
                publisher=publisher, authors=authors, library=library,
                publication_year=date, is_electronic=electronic,
                language=language
            )
            
            book_tags = [db.Tag(text=c, book=book) for c in tags.split()]
            
            session.add(book)
            session.add_all(book_tags)
    
    args = tuple()
    subcommands = [section, book]

class edit(IqraCommand):
    'edit a section or book'
    
    class section(IqraCommand):
        'edit a section'
        
        args = (
            (['old_number'], dict(type=str, help='current section number including all parent sections')),
            (['--number'], dict(type=str, default=None, help='new section number including all parent sections')),
            (['--name'], dict(type=str, default=None, help='new section name')),
            (['--abbr'], dict(type=str, default=None, help='new section abbreviation')),
        )
        
        def call(self, old_number, number, name, abbr):
            old_number = [int(c) for c in old_number.split('-')]
            
            session = self.util.get_scoped_session()
            section = self.util.get_root_section(session)
            for c in old_number:
                section = session.query(db.Section).filter(
                    db.Section.number == c, db.Section.parent == section).one()
            
            if number is not None:
                number = [int(c) for c in number.split('-')]
                new_parent = self.util.get_root_section(session)
                for c in number[0:-1]:
                    new_parent = session.query(db.Section).filter(
                        db.Section.number == c, db.Section.parent == new_parent).one()
                
                section.parent = new_parent
                section.number = number[-1]
            
            if name is not None:
                section.name = name
            
            if abbr is not None:
                section.abbr = abbr
    
    class book(IqraCommand):
        'edit a book'
        
        args = (
            (['old_number'], dict(type=str, help='old book number including all parent sections')),
            (['-n', '--number'], dict(type=str, default=None, help='new book number including all parent sections')),
            (['-t', '--title'], dict(type=str, default=None, help='new book title')),
            (['-p', '--publisher'], dict(type=str, default=None, help='new publisher')),
            (['-np', '--no-publisher'], dict(action='store_true', default=False, help='remove publisher')),
            (['-a', '--authors'], dict(type=str, default=None, help='new comma separated author names')),
            (['-l', '--library'], dict(type=str, default=None, help='new library name')),
            (['-nl', '--no-library'], dict(action='store_true', default=False, help='remove library')),
            (['-d', '--date'], dict(type=int, default=None, help='new publication year')),
            (['-nd', '--no-date'], dict(action='store_true', default=False, help='remove publication date')),
            (['-e', '--electronic'], dict(action='store_true', default=None, help='the book is an electronic book. opposite -ne')),
            (['-ne', '--not-electronic'], dict(action='store_false', dest='electronic', help='the book is a paper book. opposite -e')),
            (['--language'], dict(type=str, default=None, help='new language of the book')),
            (['--tags'], dict(type=str, default=None, help='new space seperated book tags')),
        )
        
        def call(self, old_number, number=None, title=None,
                publisher=None, no_publisher=False, authors=None, library=None,
                no_library=False, date=None, no_date=False, electronic=False,
                language=None, tags=None):
            old_number = [int(c) for c in old_number.split('-')]
            
            session = self.util.get_scoped_session()
            section = self.util.get_root_section(session)
            for c in old_number[0:-1]:
                section = session.query(db.Section).filter(
                    db.Section.number == c, db.Section.parent == section).one()
            
            old_number = old_number[-1]
            
            books = session.query(db.Book).filter_by(section=section,
                number=old_number).all()
            
            for book in books:
                if number is not None:
                    number = [int(c) for c in number.split('-')]
                
                    section = self.util.get_root_section(session)
                    for c in number[0:-1]:
                        section = session.query(db.Section).filter(
                            db.Section.number == c, db.Section.parent == section).one()
                    
                    book.section = section
                    book.number = number[-1]
                
                if title is not None:
                    book.title = title
                
                if no_publisher:
                    book.publisher = None
                elif publisher is not None:
                    obj = session.query(db.Publisher).filter(
                        db.Publisher.name == publisher).first()
                    
                    if obj is None:
                        obj = Publisher(name=publisher)
                    
                    book.publisher = obj
                
                if authors is not None:
                    authors = [c.strip() for c in authors.split(',')]
                    authors = [c for c in authors if c]
                    for idx, c in enumerate(authors):
                        obj = session.query(db.Author).filter(
                            db.Author.name == c).first()
                        if obj is None:
                            obj = db.Author(name=c)
                        
                        authors[idx] = obj
                    book.authors = authors
                
                if no_library:
                    book.library = None
                elif library is not None:
                    obj = session.query(db.Library).filter(
                        db.Library.name == library).first()
                    
                    if obj is None:
                        obj = db.Library(name=library)
                    
                    book.library = obj
                
                if no_date:
                    book.publication_year = None
                elif date is not None:
                    book.publication_year = date
                
                if electronic is not None:
                    book.is_electronic = electronic
                
                if language is not None:
                    book.language = language
                
                if tags is not None:
                    tags = [db.Tag(text=c, book=book) for c in tags.split()]
                    book.tags = tags
    
    args = tuple()
    subcommands = [section, book]

class delete(IqraCommand):
    'delete a book or section'
    
    class section(IqraCommand):
        'delete a section'
        
        args = (
            (['number'], dict(type=str, help='section number including all parent sections')),
        )
        
        def call(self, number):
            number = [int(c) for c in number.split('-')]
            
            session = self.util.get_scoped_session()
            section = self.util.get_root_section(session)
            for c in number:
                section = session.query(db.Section).filter(
                    db.Section.number == c, db.Section.parent == section).one()
            
            session.delete(section)
    
    class book(IqraCommand):
        'delete a book'
        
        args = (
            (['number'], dict(type=str, help='book number including all parent sections')),
            (['-i', '--index'], dict(type=int, default=0, help='book index (in case multiple copies of this book exists')),
        )
                
        def call(self, number, index):
            number = [int(c) for c in number.split('-')]
            
            session = self.util.get_scoped_session()
            section = self.util.get_root_section(session)
            for c in number[0:-1]:
                section = session.query(db.Section).filter(
                    db.Section.number == c, db.Section.parent == section).first()
            
            books = session.query(db.Book).filter(
                db.Book.section == section, db.Book.number == number[-1]
            ).offset(index).all()
            
            for book in books:
                session.delete(book)
    
    args = tuple()
    subcommands = [section, book]

class list_(IqraCommand):
    'list sections or books'
    name = 'list'
    
    class section(IqraCommand):
        'list sections'
        
        args = (
            (['number'], dict(type=int, default=0, nargs='?', help='max number of listed records')),
            (['-o', '--offset'], dict(type=int, default=0, help='list offset from first item')),
        )
                
        def call(self, number, offset):
            session = self.util.get_scoped_session()
            
            root_section = self.util.get_root_section(session)
            
            digits = self.util.get_section_digits()
            padding = digits * '0'
            full_number = func.substr(padding + db.Section.number.cast(String),
                -digits, digits)
            icteq = session.query(
                    db.Section.id,
                    full_number.label('number'),
                    db.Section.name,
                    db.Section.abbreviation,
                ).filter_by(
                    parent=root_section
                ).cte()
            rcteq = session.query(
                    db.Section.id,
                    icteq.c.number.concat('-').concat(full_number),
                    db.Section.name,
                    db.Section.abbreviation,
                ).join(
                    icteq, icteq.c.id == db.Section.parent_id
                )
            cteq = icteq.union(rcteq)
            
            q = session.query(cteq).order_by('number').offset(
                offset
            )
            if number > 0:
                q = q.limit(number)
            
            for pk, full_number, name, abbr in q:
                print(full_number, name, abbr)
    
    class book(IqraCommand):
        'list books'
        
        args = (
            (['number'], dict(type=int, default=10, nargs='?', help='max number of listed records')),
            (['-o', '--offset'], dict(type=int, default=0, help='list offset from first item')),
        )
                
        def call(self, number, offset):
            session = self.util.get_scoped_session()
            
            root_section = self.util.get_root_section(session)
            
            digits = self.util.get_section_digits()
            padding = digits * '0'
            full_number = func.substr(padding + db.Section.number.cast(String),
                -digits, digits)
            icteq = session.query(
                    db.Section.id,
                    full_number.label('number'),
                    db.Section.name,
                    db.Section.abbreviation
                ).filter_by(parent=root_section).cte()
            rcteq = session.query(
                    db.Section.id,
                    icteq.c.number.concat('-').concat(full_number),
                    db.Section.name,
                    db.Section.abbreviation
                ).join(
                    icteq, icteq.c.id == db.Section.parent_id
                )
            cteq = icteq.union(rcteq)
            
            subq = session.query(cteq).subquery()
            
            digits = self.util.get_book_digits()
            padding = digits * '0'
            book_number = func.substr(padding + db.Book.number.cast(String),
                -digits, digits)
            q = session.query(
                    db.Book,
                    func.count(db.Book.id),
                    subq.c.number.concat('-').concat(book_number)
                ).join(
                subq, subq.c.id == db.Book.section_id).order_by(
                subq.c.number, book_number).group_by(
                    db.Book.number, db.Book.section_id
                ).offset(offset).limit(number)
            
            for c, copies, number in q:
                if c.library is not None:
                    print(c.library.name)
                print(number, c.title, '({} copies)'.format(copies), c.language)
                print('tags:', *c.tags)
                print('authors:', *c.authors, sep=', ')
                
                publishing_info = []
                if c.publisher is not None:
                    publishing_info.append(c.publisher.name)
                if c.publication_year is not None:
                    publishing_info.append(str(c.publication_year))
                if publishing_info:
                    print(' '.join(publishing_info))
                if c.is_electronic:
                    print('electronic')
                else:
                    print('paper')
                
                print()
    
    args = tuple()
    subcommands = [section, book]

class file(IqraCommand):
    'execute commands from file'
    
    args = (
        (['name'], dict(type=str, help='file name to execute')),
    )
    
    def call(self, name):
        cmd = IqraCmd(util=self.util)
        
        with open(name, encoding='utf-8') as f:
            for c in f:
                cmd.onecmd(c)

class commit(IqraCommand):
    'commit database changes'
    
    args = tuple()
        
    def call(self):
        self.util.get_scoped_session().commit()

class rollback(IqraCommand):
    'rollback database changes'
    
    args = tuple()
        
    def call(self):
        self.util.get_scoped_session().rollback()

class echo(IqraCommand):
    'print message'
    
    args = (
        (['msg'], dict(type=str, nargs='*', help='message to print')),
    )
        
    @staticmethod
    def call(msg):
        print(*msg)

class backup(IqraCommand):
    'backup database'
    
    args = tuple()
        
    def call(self):
        self.util.backup()

class lsbackup(IqraCommand):
    'list existing database backups'
    
    args = tuple()
        
    def call(self):
        for c in self.util.list_backups():
            print(c.stem)

class restore(IqraCommand):
    'restore from backup'
    
    args = (
            (['dt'], dict(type=str, nargs='?', default=None, help='name of the backup file without extension as listed by `lsbackup` command. e.g. 20190807_112500')),
        )
    
    def call(self, dt=None):
        self.util.restore(dt)

class thumb(IqraCommand):
    'creates thumbnails for books. the created thumbnails are svg files containing the name of the book in a rectangle. if a thumbnail is found for a book, it will be ignored and no thumbnail will be generated.'
    
    args = (
        (['number'], dict(type=str, action='append', help='number of the book to generate thumb for.')),
    )
    
    def call(self, number):
        session = self.util.get_scoped_session()
        books = []
        for n in number:
            n = n.split('-')
            books.extend(self.util.get_books(session, *n).all())
        
        self._finished = 0
        self._total = len(books)
        
        for c in books:
            print(list(self.util.glob_path(
                    'res/thumbs/{}.*'.format(c.id))))
            if not len(list(self.util.glob_path(
                    'res/thumbs/{}.*'.format(c.id)))):
                self.util.generate_thumb(c)




'''
:Date: 2019-11-18
:Version: 1.0.0
:Authors:
    * Mohammad Alghafli <thebsom@gmail.com>

Iqra, a library management program.
This is the core iqra library. It also acts as a commandline interface to interact with iqra database with basic commands.
If you are looking for a GUI library management program, look for giqra which provides a user interface to interact with iqra database.

------------
Installation
------------

On windows install using pip by running the
command::
    
    pip install iqra

Or on linux::
    
    pip3 install iqra

All dependancies will be installed automatically by pip. The stickers plugin needs pycairo so you need to install it if you want to use this plugin.

-----
Usage
-----

Running the commandline interface::

    python3 -m iqra

Use -h option to know all the possible options::

    python3 -m iqra -h

You can use iqra as a library and import it in your project::

    import iqra
'''

import json
import threading
import logging
from . import db
import traceback
import appdirs
import codi
import pathlib
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
import sqlalchemy.event
import os
import cmd
import pkg_resources
import shlex
import argparse
import datetime
import shutil

try:
    import cairo
    import gi
    gi.require_version('Pango', '1.0')
    gi.require_version('PangoCairo', '1.0')
    
    from gi.repository import Pango
    from gi.repository import PangoCairo
except ImportError:
    cairo = None

__version__ = '1.0.0'

logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s: %(message)s', datefmt='%Y-%m-%d %H:%M:%S')

logger = logging.getLogger(__name__)

class IqraUtil:
    CONFIG = {
            'library_name': 'Library',
            'language': 'en',
            'backup_count': 10,
            'section_digits': 0,
            'book_digits': 0,
        }
    
    def __init__(self, config_dir=None, profile='default'):
        global _
        
        dbname = 'library_{}.db'.format(__version__.split('.')[0])
        
        config_dirs = [pathlib.Path(__file__).parent]
        if config_dir is not None:
            p = pathlib.Path(config_dir)
            config_dirs.insert(0, p / profile)
        else:
            portable_config = pathlib.Path(__file__).parent / 'iqra.cfg'
            if not portable_config.exists():
                p = pathlib.Path(appdirs.user_config_dir('iqra', 'mhghafli'))
                config_dirs.insert(0, p / profile)
        
        self.codi = codi.Codi(*config_dirs)
        
        self.new_config = False
        self.config = codi.Config()
        self.add_default_config(self.CONFIG)
        self.config.update(self.load_config())
        
        self.dbpath = self.codi.path(dbname, writable=True)
        dburi = 'sqlite:///{}'.format(self.dbpath)
        self.engine = create_engine(dburi, isolation_level=None, echo=False)
        sqlalchemy.event.listen(self.engine, 'begin', self.on_engine_begin)
        
        db.Base.metadata.create_all(self.engine)
        
        self.Sessionmaker = sessionmaker(bind=self.engine)
        self.Scopedmaker = scoped_session(sessionmaker(bind=self.engine))
        
        self.create_root_section()
        
        self.language = self.get_config('language')
        os.putenv('LANGUAGE', self.language)
    
    def create_root_section(self):
        session = self.get_session()
        
        if session.query(db.Section).filter_by(parent_id=None).count() < 1:
            section = db.Section(
                id=0, parent_id=None, number=0, name='', abbreviation='')
            session.add(section)
            session.commit()
        
        if session.query(db.Section).filter_by(parent_id=None).count() > 1:
            for c in session.query(db.Section).filter_by(section_id=None).offset(1):
                session.delete(c)
            
            session.commit()
        
        session.close()
    
    def get_root_section(self, session):
        return session.query(db.Section).get(0)
    
    def get_section(self, session, *numbers):
        section = self.get_root_section(session)
        for c in numbers:
            section = session.query(db.Section).filter_by(
                parent=section, number=c).one()
        return section
    
    def get_books(self, session, *numbers, columns=[db.Book]):
        section = self.get_section(session, *numbers[:-1])
        q = session.query(*columns).filter_by(
            section=section, number=numbers[-1])
        return q
    
    def backup(self):
        budir = self.get_path('backup', True)
        try:
            budir.mkdir(parents=True)
        except FileExistsError:
            pass
        dt = datetime.datetime.now()
        dest = budir / '{:%Y%m%d_%H%M%S}.db'.format(dt)
        
        conn = self.engine.connect()
        conn.execute('BEGIN IMMEDIATE TRANSACTION')
        shutil.copy2(str(self.dbpath), str(dest))
        conn.execute('ROLLBACK')
        conn.close()
        
        backups = self.list_backups()
        to_delete = len(backups) - self.get_config('backup_count')
        if to_delete > 0:
            to_remove = backups[:to_delete]
            for c in to_remove:
                c.unlink()
    
    def list_backups(self):
        return sorted([c for c in self.glob_path('backup/*.db') if c.is_file()])
    
    def restore(self, backup=None):
        backups = self.list_backups()
        if backup is None:
            backup = backups[-1].stem
        
        try:
            backup = [c for c in backups if c.stem == backup][0]
        except IndexError:
            raise FileNotFoundError('backup file does not exist')
        v = __version__.split('.')[0]
        dest = self.dbpath
        shutil.copy2(str(backup), str(self.dbpath))
    
    def on_engine_begin(self, conn):
        conn.execute('BEGIN')
    
    def add_default_config(self, def_conf):
        for c in def_conf:
            try:
                self.config.get_default(c)
            except KeyError:
                self.config.set_default(c, def_conf[c])
    
    def load_config(self):
        try:
            return json.loads(self.codi.read('iqra.cfg'))
        except Exception as e:
            if type(e) is not FileNotFoundError:
                print('error loading config')
                print(e)
                print(traceback.format_exc())
                print('will create new config')
            
            self.new_config = True
            self.codi.write('iqra.cfg', json.dumps({}))
            
            return {}
    
    def dump_config(self):
        try:
            self.codi.write('iqra.cfg', json.dumps(self.config))
        except Exception as e:
            logger.warning('could not write config file. Error: {} `{}`'.format(
                type(e).__name__, e))
    
    def get_config_names(self):
        return sorted(self.config)
    
    def get_config(self, name):
        return self.config[name]
    
    def set_config(self, name, value):
        self.config[name] = value
    
    def get_path(self, fpath, writable=False):
        return self.codi.path(fpath, writable)
    
    def glob_path(self, pattern):
        return self.codi.glob(pattern)
    
    def get_session(self, *args, **kwargs):
        return self.Sessionmaker(*args, **kwargs)
    
    def get_scoped_session(self, *args, **kwargs):
        session = self.Scopedmaker(*args, **kwargs)
        return session
    
    def get_commands(self):
        return pkg_resources.iter_entry_points('iqra.commands')
    
    def get_plugins(self, name=None):
        plugins = [
            c.load() for c in pkg_resources.iter_entry_points(
                'iqra.plugins', name)
        ]
        
        for c in plugins:
            try:
                self.add_default_config(c.CONFIG)
            except AttributeError:
                pass
        return plugins
    
    def get_section_digits(self):
        digits = self.get_config('section_digits')
        if digits <= 0:
            digits = self.find_section_digits()
        
        return digits
    
    def find_section_digits(self):
        session = self.get_scoped_session()
        q = session.query(db.Section.number).order_by(
            db.Section.number.desc())
        obj = q.first()
        if obj is not None:
            digits = len(str(obj[0]))
        else:
            digits = 1
        
        return digits
    
    def get_book_digits(self):
        digits = self.get_config('book_digits')
        if digits <= 0:
            digits = self.find_book_digits()
        
        return digits
    
    def find_book_digits(self):
        session = self.get_scoped_session()
        q = session.query(db.Book.number).order_by(db.Book.number.desc())
        obj = q.first()
        if obj is not None:
            digits = len(str(obj[0]))
        else:
            digits = 1
        
        return digits
    
    def get_thumbs(self, id):
        return self.glob_path('res/thumbs/{}.*'.format(id))
    
    def delete_thumb(self, id):
        for c in self.get_thumbs(id):
            c.unlink()
    
    def set_thumb(self, id, thumb):
        thumb = pathlib.Path(thumb)
        thumbs_dir = self.get_path('res/thumbs/', writable=True).mkdir(
            parents=True, exist_ok=True)
        thumb_path = self.get_path(
            'res/thumbs/{}{}'.format(id, thumb.suffix), writable=True)
        if thumb != thumb_path:
            shutil.copy(str(thumb), str(thumb_path))
        
        to_delete = [c for c in self.get_thumbs(id) if c != thumb_path]
        for c in to_delete:
            c.unlink()
    
    def add_electronic_file(self, id, path):
        print('adding', path)
        path = pathlib.Path(path)
        e_dir = self.get_path('res/electronic/', writable=True).mkdir(
            parents=True, exist_ok=True)
        e_path = self.get_path(
            'res/electronic/{}{}'.format(id, path.suffix), writable=True)
        if path != e_path:
            shutil.copy(str(path), str(e_path))
    
    def delete_electronic_file(self, id, ext):
        e_path = self.get_path(
            'res/electronic/{}{}'.format(id, ext), writable=True)
        e_path.unlink()
    
    def get_electronic_files(self, id):
        return self.glob_path('res/electronic/{}.*'.format(id))
    
    def generate_thumb(self, book):
        if cairo is None:
            raise RuntimeError('cairo library is missing. must install it.')
        
        H = 128
        W = 128
        LINE_WIDTH = 4
        TEMPLATE = '{.title}'
        
        if type(book) is int:
            book = self.get_scoped_session().query(db.Book).get(book)
        
        path = self.get_path('res/thumbs/{}.svg'.format(book.id),
            writable=True)
        path.parent.mkdir(parents=True, exist_ok=True)
        
        surface = cairo.SVGSurface(str(path), W, H)
        
        ctx = cairo.Context(surface)
        
        ctx.set_source_rgb(1, 1, 1)
        ctx.rectangle(0, 0, W, H)
        ctx.fill()
        
        ctx.set_line_width(LINE_WIDTH)
        x = y = LINE_WIDTH / 2
        w = W - LINE_WIDTH
        h = H - LINE_WIDTH
        ctx.rectangle(x, y, w, h)
        ctx.set_source_rgb(0, 0, 0)
        ctx.stroke_preserve()
        ctx.clip()
        
        lout = PangoCairo.create_layout(ctx)
        lout.set_font_description(Pango.FontDescription('Sans 10'))
        lout.set_width(Pango.SCALE * w)
        lout.set_alignment(Pango.Alignment.CENTER)
        
        text = TEMPLATE.format(book)
        lout.set_markup(text, -1)
        extent = lout.get_size()
        xtrans = x
        ytrans = y - extent.height / Pango.SCALE / 2 + h / 2
        
        ctx.translate(xtrans, ytrans)
        PangoCairo.update_layout(ctx, lout)
        PangoCairo.show_layout(ctx, lout)
        
        surface.finish()
        
        return path

class IqraCmd(cmd.Cmd):
    prompt = '>>> '
    
    def __init__(self, util=None, plugins=None, commands=None, *args, **kwargs):
        if util is None:
            util = IqraUtil()
        
        if plugins is None:
            plugins = util.get_plugins()
        
        if commands is None:
            commands = [c.load() for c in util.get_commands()]
        
        self.util = util
        self.session = util.get_session()
        
        self.parser = argparse.ArgumentParser(prog='', add_help=False,
            description='iqra, a library management program')
        
        parsers = self.parser.add_subparsers()
        
        self.commands = set()
        
        self.add_plugins(parsers, plugins)
        self.add_commands(parsers, commands, top_commands=True)
        
        if not hasattr(type(self), 'intro'):
            type(self).intro = 'welcome to {}'.format(
                self.util.get_config('library_name')
            )
        
        super().__init__(*args, **kwargs)
    
    def do_exit(self, arg):
        '''end command line program.'''
        return True
    
    def do_EOF(self, arg):
        '''end command line program.'''
        print()
        return self.do_exit(arg)
    
    def cmdloop(self, *args, **kwargs):
        while True:
            try:
                super().cmdloop()
                break
            except Exception as e:
                print(e)
                print(traceback.format_exc())
    
    def emptyline(self):
        pass
    
    def default(self, arg):
        self.execute_line(arg)
    
    def execute_line(self, args):
        if type(args) is str:
            args = shlex.split(args)
        try:
            ns = vars(self.parser.parse_args(args))
            ns.pop('func')(**ns)
        except SystemExit:
            pass
    
    def add_plugins(self, parsers, plugins):
        for c in plugins:
            if hasattr(c, 'name'):
                name = c.name
            else:
                name = c.__name__
            parser = parsers.add_parser(name, prog=name, description=c.__doc__, **c.parser_args)
            
            cmd_method = 'do_{}'.format(name)
            cmd_help = 'help_{}'.format(name)
            self.commands.add(cmd_method)
            self.commands.add(cmd_help)
            setattr(self, cmd_help, parser.print_help)
            
            caller = PluginCaller(c, util=self.util)
            parser.set_defaults(func=caller.call)
            
            for args in c.args:
                parser.add_argument(*args[0], **args[1])
    
    def add_commands(self, parsers, commands, top_commands=False):
        for c in commands:
            if hasattr(c, 'name'):
                name = c.name
            else:
                name = c.__name__
            parser = parsers.add_parser(name, prog=name, description=c.__doc__, **c.parser_args)
            
            if top_commands:
                cmd_method = 'do_{}'.format(name)
                cmd_help = 'help_{}'.format(name)
                self.commands.add(cmd_method)
                self.commands.add(cmd_help)
                setattr(self, cmd_help, parser.print_help)
            
            parser.set_defaults(func=c(util=self.util).call)
            
            for args in c.args:
                parser.add_argument(*args[0], **args[1])
            
            if hasattr(c, 'subcommands'):
                subparsers = parser.add_subparsers()
                self.add_commands(subparsers, c.subcommands)
    
    def get_names(self):
        return sorted(dir(type(self)) + list(self.commands))

class PluginCaller:
    def __init__(self, plugin_cls, util=None):
        if util is None:
            util = IqraUtil()
        
        self.util = util
        self.plugin = plugin_cls(util=self.util, close=False)
        self.stopped = threading.Event()
        self.stopped.set()
    
    def call(self, *args, **kwargs):
        self.stopped.clear()
        cancelled = False
        t = threading.Thread(target=self.print_info, daemon=True)
        t.start()
        try:
            self.plugin.call(*args, **kwargs)
        except KeyboardInterrupt:
            cancelled = True
        finally:
            self.stopped.set()
        
        t.join(0.1)
        if cancelled:
            print('cancelled')
        else:
            print()
    
    def print_info(self):
        message_length = 0
        while not self.stopped.wait(timeout=0.1):
            message = self.get_info()
            print(message.ljust(message_length), end='', flush=True)
            message_length = len(message)
        
        message = self.get_info()
        print(message.ljust(message_length), end='', flush=True)
    
    def get_info(self):
        finished = self.plugin.finished
        total = self.plugin.total
        digits = str(len(str(total)))
        message = self.plugin.message
        base_str = '\r{:><digits>}/{}'.replace('<digits>', digits)
        progress = base_str.format(finished,total)
        return '{} {}'.format(progress, message)




from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Table, Column, Integer, Text, String, DateTime, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.orm import Query
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy import CheckConstraint, UniqueConstraint
import datetime
from .. import db

Base = declarative_base()

class Section(Base):
    __tablename__ = 'section'
    __gui_metadata__ = {
            'combo': {'columns': ('id', 'number', 'name'), 'cast': (String,), 'extra_fields': {'text':'{number:0<section_digits>}-{name}'}},
            'dialog': ('number', 'name', 'abbreviation')
        }
    __query_metadata__ = {
            'order': ('number',)
        }
    __export_headers__ = [
            'Number',
            'Name',
            'Abbreviation',
        ]
    __import_class__ = db.Section
    
    id = Column(Integer, primary_key=True, index=True)
    number = Column(Integer, CheckConstraint('number >= 0'), unique=True, nullable=False, index=True)
    name = Column(String, CheckConstraint("name <> ''"), index=True, nullable=False)
    abbreviation = Column(String, default='', nullable=False)
    
    subsections = relationship("Subsection", back_populates="section", cascade='all, delete-orphan')
    books = relationship('Book',
        secondary="join(Subsection, Book)",
        primaryjoin="Section.id == Subsection.section_id",
        secondaryjoin="Subsection.id == Book.subsection_id",
        viewonly=True,
    )
    
    def __repr__(self):
        return "<Section {0.number:02} {0.name}>".format(self)
    
    def exporter(self):
        return [self.number, self.name, self.abbreviation]
    
    @classmethod
    def importer(class_, *args):
        args = default_args(args, [None, None, ''], class_.__export_headers__)
        return {'number': int(args[0]), 'name': args[1], 'abbreviation': args[2]}
    
    @classmethod
    def validator(class_, session, values):
        q = session.query(class_).filter(class_.number == values['number'])
        if q.count() > 0:
            raise ValueError('Section exists')
    
    def importdb(self, session):
        root_section = session.query(db.Section).get(0)
        return [db.Section(number=self.number, name=self.name,
            abbreviation=self.abbreviation, parent=root_section)]

class Subsection(Base):
    __tablename__ = 'subsection'
    __table_args__ = (UniqueConstraint('section_id', 'number'),)
    __gui_metadata__ = {
            'combo': {'columns': ('id', 'number', 'name', 'section_id'), 'cast': (String,), 'extra_fields': {'text':'{number:0<subsection_digits>}-{name}'}, 'chain': 'section'},
            'dialog': ('section', 'number', 'name', 'abbreviation')
        }
    __query_metadata__ = {
            'order': ('number',)
        }
    
    __export_headers__ = [
            'Number',
            'Name',
            'Abbreviation',
        ]
    __import_class__ = db.Section
    
    id = Column(Integer, primary_key=True, index=True)
    section_id = Column(Integer, ForeignKey('section.id'), nullable=False)
    number = Column(Integer, CheckConstraint('number >= 0'), nullable=False, index=True)
    name = Column(String, CheckConstraint("name <> ''"), index=True, nullable=False)
    abbreviation = Column(String, default='', nullable=False)
    
    section = relationship("Section", back_populates="subsections")
    books = relationship("Book", back_populates="subsection")
    
    def __repr__(self):
        return "<Subsection {0.section.number:02}-{0.number:02} {0.section.name}-{0.name}>".format(self)
    
    def exporter(self):
        number = '{0.section.number}-{0.number}'.format(self)
        return [number, self.name, self.abbreviation]
    
    @classmethod
    def importer(class_, *args):
        args = default_args(args, [None, None, ''], class_.__export_headers__)
        section_number, number = args[0].split('-')
        section = {'query': Query(Section).filter(
            Section.number == int(section_number))}
        return {'section': section,
            'number': int(number), 'name': args[1], 'abbreviation': args[2]}
    
    @classmethod
    def validator(class_, session, values):
        q = session.query(class_).filter(
            Subsection.section_id == values['section'].id,
            class_.number == values['number'])
        
        if q.count() > 0:
            raise ValueError('Subsection exists')
    
    def importdb(self, session):
        root_section = session.query(db.Section).get(0)
        parent = session.query(db.Section).filter(
            db.Section.parent == root_section,
            db.Section.number == self.section.number
        ).one()
        return [db.Section(number=self.number, name=self.name,
            abbreviation=self.abbreviation, parent=parent)]

book_author = Table('book_author', Base.metadata,
    Column('id', Integer, primary_key=True, index=True),
    Column('book_id', Integer, ForeignKey('book.id', ondelete='cascade'), nullable=False),
    Column('author_id', Integer, ForeignKey('author.id', ondelete='cascade'), nullable=False),
    UniqueConstraint('book_id', 'author_id')
)

class BookAuthor(Base):
    __table__ = book_author
    __mapper_args__ = {
            'confirm_deleted_rows': False
        }
    id = book_author.c.id
    book_id = book_author.c.book_id
    author_id = book_author.c.author_id
    
    book = relationship("Book")
    author = relationship("Author")

class Author(Base):
    __tablename__ = 'author'
    __gui_metadata__ = {
            'combo': {'columns': ('id', 'name'), 'cast': (String,), 'labels': ('id',), 'text': 'name', 'with_entry': True},
            'dialog': ('name', 'about',)
        }
    __query_metadata__ = {
            'order': ('name',)
        }
    __export_headers__ = [
            'Name',
            'About',
        ]
    __import_class__ = db.Author
    
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, CheckConstraint("name <> ''"), unique=True, index=True, nullable=False)
    about = Column(Text, nullable=False, default='')
    
    books = relationship("Book", secondary=book_author, back_populates="authors")
    
    def exporter(self):
        return [self.name, self.about]
    
    @classmethod
    def importer(class_, *args):
        args = default_args(args, [None, ''], class_.__export_headers__)
        return {'name': args[0], 'about': args[1]}
    
    @classmethod
    def validator(class_, session, values):
        q = session.query(class_).filter(class_.name == values['name'])
        if q.count() > 0:
            raise ValueError('Author exists')
    
    def importdb(self, session):
        return [db.Author(name=self.name, about=self.about)]

class Publisher(Base):
    __tablename__ = 'publisher'
    __gui_metadata__ = {
            'combo': {'columns': ('id', 'name'), 'cast': (String,), 'labels': ('id',), 'text': 'name', 'with_entry': True},
            'dialog': ('name', 'about',)
        }
    __query_metadata__ = {
            'order': ('name',)
        }
    __export_headers__ = [
            'Name',
            'About',
        ]
    __import_class__ = db.Publisher
    
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, CheckConstraint("name <> ''"), unique=True, index=True, nullable=False)
    about = Column(Text, nullable=False, default='')
    
    books = relationship("Book", back_populates="publisher")
    
    def exporter(self):
        return [self.name, self.about]
    
    @classmethod
    def importer(class_, *args):
        args = default_args(args, [None, ''], class_.__export_headers__)
        return {'name': args[0], 'about': args[1]}
    
    @classmethod
    def validator(class_, session, values):
        q = session.query(class_).filter(class_.name == values['name'])
        if q.count() > 0:
            raise ValueError('Publisher exists')
    
    def importdb(self, session):
        return [db.Publisher(name=self.name, about=self.about)]

class Location(Base):
    __tablename__ = 'location'
    __gui_metadata__ = {
            'combo': {'columns': ('id', 'name'), 'cast': (String,), 'labels': ('id',), 'text': 'name'},
            'dialog': ('name',)
        }
    __query_metadata__ = {
            'order': ('name',)
        }
    __export_headers__ = [
            'Name',
        ]
    __import_class__ = db.Library
    
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, CheckConstraint("name <> ''"), unique=True, index=True, nullable=False)
    
    books = relationship("Book", back_populates="location")
    
    def exporter(self):
        return [self.name]
    
    @classmethod
    def importer(class_, *args):
        args = default_args(args, [None], class_.__export_headers__)
        return {'name': args[0]}
    
    @classmethod
    def validator(class_, session, values):
        q = session.query(class_).filter(class_.name == values['name'])
        if q.count() > 0:
            raise ValueError('Location exists')
    
    def importdb(self, session):
        return [db.Library(name=self.name)]

class Book(Base):
    __tablename__ = 'book'
    __table_args__ = (UniqueConstraint('location_id', 'subsection_id', 'number'),)
    __gui_metadata__ = {
            'dialog': ('subsection', 'number', 'title', 'authors', 'publisher', 'location', 'copies'),
            'combo': {'columns': ('id', 'number', 'title', 'subsection_id'), 'cast': (String,), 'extra_fields': {'text':'{number:0<book_digits>}-{title}'}, 'chain': 'subsection'},
        }
    __many_relationships__ = {
            'authors': (BookAuthor, 'book_id', 'author_id')
        }
    __export_headers__ = [
            'Book Number',
            'Title',
            'Authors',
            'Publisher',
            'Location',
            'Copies',
        ]
    __import_class__ = db.Book
    
    id = Column(Integer, primary_key=True, index=True)
    section_id = association_proxy('subsection','section_id')
    subsection_id = Column(Integer, ForeignKey('subsection.id'), nullable=False)
    number = Column(Integer, CheckConstraint('number > 0'), nullable=False, index=True)
    title = Column(String, CheckConstraint("title <> ''"), index=True, nullable=False)
    publisher_id = Column(Integer, ForeignKey('publisher.id'))
    location_id = Column(Integer, ForeignKey('location.id'))
    copies = Column(Integer, CheckConstraint('copies >= 0'), nullable=False, default=1)
    
    section = association_proxy('subsection','section')
    subsection = relationship("Subsection", back_populates="books")
    authors = relationship("Author", secondary=book_author, back_populates="books")
    publisher = relationship("Publisher", back_populates="books")
    location = relationship("Location", back_populates="books")
    borrows = relationship("Borrow", cascade='all, delete-orphan')
    
    def __repr__(self):
        return "<Book {0.section.number:02}-{0.subsection.number:02}-{0.number:04} {0.section.name}-{0.subsection.name}-{0.title}>".format(self)
    
    def exporter(self):
        book_number = '{}-{}-{}'.format(self.section.number, self.subsection.number, self.number)
        authors = ','.join([c.name for c in self.authors])
        if self.publisher is None:
            publisher = ''
        else:
            publisher = self.publisher.name
        if self.location is None:
            location = ''
        else:
            location = self.location.name
        return [
                book_number,
                self.title,
                authors,
                publisher,
                location,
                self.copies,
            ]
    
    @classmethod
    def importer(class_, *args):
        args = default_args(
            args, [None, None, '', '', '', 1], class_.__export_headers__)
        section_number, subsection_number, book_number = args[0].split('-')
        subsection = {'query': Query(Subsection).join(Section).filter(Section.number == int(section_number), Subsection.number == int(subsection_number))}
        if args[3]:
            publisher = {
                    'query': Query(Publisher).filter_by(name=args[3]),
                    'add': Publisher(name=args[3]),
                }
        else:
            publisher = None
        if args[4]:
            location = {
                    'query': Query(Location).filter_by(name=args[4]),
                    'add': Location(name=args[4])
                }
        else:
            location = None
        author_names = args[2].split(',')
        author_names = [name.strip() for name in author_names]
        author_names = [name for name in author_names if name]
        author_filters = [Author.name == name for name in author_names]
        author_adds = [Author(name=name) for name in author_names]
        return {
                'subsection': subsection,
                'number': int(book_number),
                'title': args[1],
                'publisher': publisher,
                'location': location,
                'copies': int(args[5]),
                BookAuthor: {
                        'rel_prop': 'authors',
                        'rel_query': Query(Author),
                        'filters': author_filters,
                        'adds': author_adds,
                    },
        }
    
    @classmethod
    def validator(class_, session, values):
        if values['location'] is not None:
            location_id = values['location'].id
        else:
            location_id = None
        q = session.query(class_).filter(
            class_.number == values['number'],
            class_.subsection_id == values['subsection'].id,
            class_.location_id == location_id)
        
        if values['number'] <= 0:
            raise ValueError('Invalid book number')
        elif values['copies'] < 0:
            raise ValueError('Invalid copies')
        elif q.count() > 0:
            raise ValueError('Book exists')
    
    def importdb(self, session):
        root_section = session.query(db.Section).get(0)
        section = session.query(db.Section).filter(
            db.Section.parent == root_section,
            db.Section.number == self.section.number
        ).one()
        subsection = session.query(db.Section).filter(
            db.Section.parent == section,
            db.Section.number == self.subsection.number
        ).one()
        if self.publisher is not None:
            publisher = session.query(db.Publisher).filter(
                db.Publisher.name == self.publisher.name).one()
        else:
            publisher = None
        author_names = [c.name for c in self.authors]
        if author_names:
            authors = session.query(db.Author).filter(
                db.Author.name.in_(author_names)).all()
        else:
            authors = []
        if self.location is not None:
            library = session.query(db.Library).filter(
                db.Library.name == self.location.name).one()
        else:
            library = None
        
        books = [db.Book(section=subsection, number=self.number,
            title=self.title, authors=authors, publisher=publisher,
            library=library, is_electronic=False, language=''
        ) for c in range(self.copies)]
        return books
            

class Borrow(Base):
    __tablename__ = 'borrow'
    __gui_metadata__ = {
            'dialog': ('book', 'borrower', 'contact'),
        }
    __export_headers__ = [
            'Book number',
            'Borrower',
            'Contact',
            'Date',
            'Return date',
        ]
    __import_class__ = db.Borrow
    
    id = Column(Integer, primary_key=True, index=True)
    book_id = Column(Integer, ForeignKey('book.id'), nullable=False)
    borrower = Column(String, CheckConstraint("borrower <> ''"), index=True, nullable=False)
    contact = Column(String, CheckConstraint("contact not like '%[^0-9]%' AND contact <> ''"), index=True, nullable=False)
    
    book = relationship("Book", back_populates='borrows')
    
    date = Column(DateTime, index=True, nullable=False, default=datetime.datetime.now())
    return_date = Column(DateTime, index=True)
    
    def return_(self):
        self.return_date = datetime.datetime.now()
    
    def exporter(self):
        book_number = '{}-{}-{}'.format(self.book.section.number, self.book.subsection.number, self.book.number)
        date = '{:%Y-%m-%d}'.format(self.date)
        if self.return_date is None:
            return_date = ''
        else:
            return_date = '{:%Y-%m-%d}'.format(self.return_date)
        
        return [
                book_number,
                self.borrower,
                self.contact,
                date,
                return_date
            ]
    
    @classmethod
    def importer(class_, *args):
        args = default_args(
            args, [None, None, None, None, ''], class_.__export_headers__)
        section_number, subsection_number, book_number = [int(c) for c in args[0].split('-')]
        book_id = {
                'query': Query(
                        Book
                    ).join(
                        Subsection
                    ).join(
                        Section, Section.id == Subsection.section_id
                    ).filter(
                        Section.number == section_number,
                        Subsection.number == subsection_number,
                        Book.number == book_number
                    )
            }
        
        date = datetime.datetime.strptime(args[3], '%Y-%m-%d')
        if args[4] != '':
            return_date = datetime.datetime.strptime(args[4], '%Y-%m-%d')
        else:
            return_date = None
        return {
                'book_id': book_id,
                'borrower': args[1],
                'contact': args[2],
                'date': date,
                'return_date': return_date,
            }
    
    def importdb(self, session):
        root_section = session.query(db.Section).get(0)
        section = session.query(db.Section).filter(parent=root_section,
            number=self.book.section.number).one()
        subsection = session.query(db.Section).filter(parent=section,
            number=self.book.subsection.number).one()
        book = session.query(db.Book).filter(section=subsection,
            number=self.book.number)
        
        return [db.Borrow(book=book, borrower=self.borrower,
            contact=self.contact, date=self.date, return_date=self.return_date)]

def default_args(args, defaults, headers=[]):
    args = list(args)
    for c in range(len(defaults)):
        if len(args) < c:
            if defaults[c] is None:
                raise ValueError('`{}` missing'.format(headers[c]))
            else:
                args.append(defaults[c])
        elif args[c] == '':
            if defaults[c] is None:
                raise ValueError('`{}` missing'.format(headers[c]))
            else:
                args[c] = defaults[c]
    
    return args

class Config(Base):
    __tablename__ = 'config'
    
    id = Column(Integer, primary_key=True, index=True)
    name = Column(String, CheckConstraint("name <> ''"), index=True, nullable=False)
    value = Column(String)

CLASSES = (
        Location,
        Author,
        Publisher,
        Section,
        Subsection,
        Book,
        Borrow,
    )




import iqra
import argparse

def main():
    parser = argparse.ArgumentParser(
        description='iqra, a library management program')
    parser.add_argument('-c', '--commit', action='store_true', default=False,
        help='commit before exit. opposite of -nc')
    parser.add_argument('-nc', '--no-commit', action='store_false',
        dest='commit',
        help='do not commit before exit. opposite of -nc. default')
    parser.add_argument('-p', '--profile', default='default',
        help='profile to open. each profile has its own database and preferences. the default profile is named `default`')
    parser.add_argument('-d', '--directory', default=None,
        help='root config dir where profiles exist. defaults to a directory named `iqra` in the usual configuration path in the system (in linux, that would be `~/.config/iqra`)')
    parser.add_argument('command', nargs=argparse.REMAINDER,
        help='command to be executed')
    
    ns = parser.parse_args()
    
    iqra_util = iqra.IqraUtil(config_dir=ns.directory, profile=ns.profile)
    
    if ns.command:
        iqra.IqraCmd(util=iqra_util).execute_line(ns.command)
    else:
        while True:
            try:
                iqra.IqraCmd(util=iqra_util).cmdloop()
                break
            except KeyboardInterrupt:
                print()
    
    if ns.commit:
        print('commiting before exit')
        iqra_util.get_scoped_session().commit()

if __name__ == '__main__':
    main()




from . import IqraUtil

class IqraCommand:
    'iqra command base class. should be inherited'
    
    parser_args = {}
    
    args = tuple()
    
    def __init__(self, util=None):
        if util is None:
            util = IqraUtil()
        
        self.util = util
    
    @classmethod
    def call(cls, **kwargs):
        cls.help()
    
    @classmethod
    def help(cls):
        print('specify a subcommand:')
        for c in cls.subcommands:
            if hasattr(c, 'description'):
                print('    {}: {}'.format(c.__name__, c.description))
            else:
                print('    {}'.format(c.__name__))

class IqraPlugin:
    'iqra plugin base class. should be inherited'
    
    parser_args = {}
    
    def __init__(self, util=None, close=True, start=None, step=None, end=None):
        if util is None:
            util = IqraUtil()
        
        self.util = util
        self.close = close
        self._finished = -1
        self._total = -1
        self._message = ''
        self._stop = False
        self.start = start
        self.step = step
        self.end = end
        
        if self.step is not None:
            self.next = self.callback_next
        else:
            self.next = self.normal_next
    
    @property
    def finished(self):
        return self._finished
    
    def callback_next(self):
        self._finished += 1
        if self.step is not None:
            self.step(self)
    
    def normal_next(self):
        self._finished += 1
    
    @property
    def total(self):
        return self._total
    
    @property
    def message(self):
        return self._message
    
    @property
    def ratio(self):
        if self.total <= 0:
            return -1
        elif self.total == self.finished:
            return 1
        else:
            return self.finished / self.total
    
    @property
    def percentage(self):
        return 100 * self.ratio
    
    def stop(self):
        self._stop = True
    
    def call(self, *args, **kwargs):
        if self.start is not None:
            self.start(self)
        session = self.util.get_scoped_session()
        session.begin_nested()
        try:
            self.do_call(*args, **kwargs)
            session.commit()
            if self.close:
                session.commit()
            if self.end is not None:
                self.end(self)
        except Exception:
            raise
        finally:
            session.close()
    
    def do_call(self, *args, **kwargs):
        raise NotImplemented('this method should be overridden in a subclass.')




'''
copied and modified from https://github.com/pypa/sampleproject
'''

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import re

long_description = open('README.txt').read()

version = '1.0.0'

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='iqra',
    version=version,
    description='iqra, a library management program.',
    long_description=long_description,
    url='https://github.com/alghafli/iqra',
    author='Mohammad Alghafli',
    author_email='thebsom@gmail.com',
    keywords='library management program',
    packages=find_packages(),
    install_requires=['sqlalchemy', 'appdirs', 'codi'],
    extras_require={
        'cairo': ['pycairo'],
    },
    include_package_data=True,
    entry_points={
        'iqra.commands': [
            'new=iqra.commands:new',
            'edit=iqra.commands:edit',
            'delete=iqra.commands:delete',
            'list=iqra.commands:list_',
            'file=iqra.commands:file',
            'commit=iqra.commands:commit',
            'rollback=iqra.commands:rollback',
            'echo=iqra.commands:echo',
            'backup=iqra.commands:backup',
            'lsbackup=iqra.commands:lsbackup',
            'restore=iqra.commands:restore',
            'thumb=iqra.commands:thumb [cairo]',
        ],
        'iqra.plugins': [
            'importdb=iqra.plugins:importdb',
            'exportdb=iqra.plugins:exportdb',
            'stickers=iqra.plugins:stickers [cairo]',
        ]
    },

    #project_urls={
    #    'Bug Reports': 'https://github.com/pypa/sampleproject/issues',
    #    'Funding': 'https://donate.pypi.org',
    #    'Say Thanks!': 'http://saythanks.io/to/example',
    #    'Source': 'https://github.com/pypa/sampleproject/',
    #},
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        # 'Intended Audience :: Developers',

        # Pick your license as you wish
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.5',
    ],
)



import sys
import csv
from .iqra_plugins import IqraPlugin
import importlib
import pathlib
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy import String
from sqlalchemy import func, text
from . import db
import argparse
try:
    import cairo
    import gi
    gi.require_version('Pango', '1.0')
    gi.require_version('PangoCairo', '1.0')
    
    from gi.repository import Pango
    from gi.repository import PangoCairo
except ImportError:
    cairo = None

class exportdb(IqraPlugin):
    'exports data to a csv file'
    
    args = (
        (['name'], dict(type=str, help='file name to export to')),
        (['-nl', '--no-library'], dict(action='append_const', const=db.Library, default=list(), dest='exclude', help='do not include libraries in the exported file')),
        (['-na', '--no-author'], dict(action='append_const', const=db.Author, dest='exclude', help='do not include authors in the exported file')),
        (['-np', '--no-publisher'], dict(action='append_const', const=db.Publisher, dest='exclude', help='do not include publishers in the exported file')),
        (['-ns', '--no-section'], dict(action='append_const', const=db.Section, dest='exclude', help='do not include sections in the exported file')),
        (['-nb', '--no-book'], dict(action='append_const', const=db.Book, dest='exclude', help='do not include books in the exported file')),
        (['--no-borrow'], dict(action='append_const', const=db.Borrow, dest='exclude', help='do not include borrows in the exported file')),
    )
    
    def do_call(self, name, exclude=tuple()):
        session = self.util.get_scoped_session()
        
        try:
            with open(name, 'w', encoding='utf-8') as f:
                writer = csv.writer(f, delimiter=',')
                classes = {
                    c: session.query(c) for c in db.CLASSES if c not in exclude
                }
                if db.Section in classes:
                    classes[db.Section] = classes[db.Section].filter(
                        db.Section.parent_id != None)
                
                classes = {c: classes[c] for c in classes if classes[c].count()}
                self._finished = 0
                self._total = sum([classes[c].count() for c in classes])
                
                for c in classes:
                    try:
                        writer.writerow([''] + c.__export_headers__)
                    except AttributeError:
                        pass
                    
                    q = classes[c]
                    self._message = 'exporting {}'.format(c.__name__)
                    for d in q:
                        row = [type(d).__name__.casefold()]
                        row.extend(iter(d))
                        writer.writerow(row)
                        self.next()
                        if self._stop:
                            break
                    
                    if self._stop:
                        break
        except Exception as e:
            self._message = str(e)
            raise
        
        self._message = 'done'

class importdb(IqraPlugin):
    'imports data from an older version iqra database or from a csv file'
    
    args = (
        (['name'], dict(type=str, help='file name to import. if the extension is `db` or `-v` argument is given, will guess the file to be a database file. otherwise, will guess the file to be a csv file.')),
        (['-v', '--version'], dict(type=int, nargs='?', default=None, help='database version. if ommitted it will be guessed from file name (e.g. for `library_0.db`, version will be guessed as 0).')),
        (['-nl', '--no-library'], dict(action='append_const', const=db.Library, default=list(), dest='exclude', help='do not include libraries in the exported file')),
        (['-na', '--no-author'], dict(action='append_const', const=db.Author, dest='exclude', help='do not include authors in the exported file')),
        (['-np', '--no-publisher'], dict(action='append_const', const=db.Publisher, dest='exclude', help='do not include publishers in the exported file')),
        (['-ns', '--no-section'], dict(action='append_const', const=db.Section, dest='exclude', help='do not include sections in the exported file')),
        (['-nb', '--no-book'], dict(action='append_const', const=db.Book, dest='exclude', help='do not include books in the exported file')),
        (['--no-borrow'], dict(action='append_const', const=db.Borrow, dest='exclude', help='do not include borrows in the exported file')),
    )
    
    def do_call(self, name, version=None, exclude=[]):
        try:
            if version is not None or name.endswith('.db'):
                self.import_db(name, version, exclude)
            else:
                self.import_csv(name, exclude)
        except Exception as e:
            self._message = str(e)
            raise
    
    def import_csv(self, name, exclude=[]):
        tables = {c.__name__.casefold(): c for c in db.CLASSES if c not in exclude}
        session = self.util.get_scoped_session()
        with open(name, encoding='utf-8') as f:
            reader = csv.reader(f, delimiter=',')
            rows = tuple(reader)
            self._finished = 0
            self._total = len(rows)
            
            self._message = 'importing {}'.format(name)
            
            for c in rows:
                table = c[0].casefold()
                if table in tables:
                    session.add(tables[table].from_values(session, *c[1:]))
                self.next()
                if self._stop:
                    break
        
        self._message = 'done'
    
    def import_db(self, name, version=None, exclude=[]):
        if version is None:
            version = int(pathlib.Path(name).stem.split('_')[-1])
        
        uri = 'sqlite:///{}'.format(name)
        
        import_name = '{}.importers.db_{}'.format(__package__, version)
        
        old_db = importlib.import_module(import_name)
        
        engine = create_engine(uri)
        old_db.Base.metadata.create_all(engine)
        
        session = self.util.get_scoped_session()
        old_session = sessionmaker(bind=engine)()
        
        classes = [c for c in old_db.CLASSES if c.__import_class__ not in exclude]
        
        count = sum([old_session.query(cls).count() for cls in classes])
        self._finished = 0
        self._total = count
        try:
            for cls in classes:
                self._message = 'importing {}'.format(cls.__name__)
                for n, c in enumerate(old_session.query(cls)):
                    session.add_all(c.importdb(session))
                    self.next()
                    if self._stop:
                        break
                if self._stop:
                    session.rollback()
                    break
        finally:
            old_session.close()
        
        self._message = 'done'

class stickers(IqraPlugin):
    CONFIG = {
            'stickers_page_width': 210,
            'stickers_page_height': 297,
            'stickers_top_margin': 1.0,
            'stickers_left_margin': 1.0,
            'stickers_voffset': 0.0,
            'stickers_hoffset': 0.0,
            'stickers_columns': 4,
            'stickers_rows': 10,
            'stickers_row_sep': 1.0,
            'stickers_col_sep': 1.0,
            'stickers_border_width': 1.0,
            'stickers_border_color': [0, 0, 0],
            'stickers_font': 'Droid Arabic Kufi 10',
            'stickers_text_color': [0, 0, 0],
            'stickers_book_template': '{book.library.name}\n'
                                      '{book.full_number}\n'
                                      '{book.section.full_abbreviation}',
            'stickers_section_template': '{section.name}\n'
                                         '{section.full_number}',
            'stickers_library_template': '{library.name}',
        }
    
    args = (
        (['-f', '--file'], dict(type=str, default='-', help='file name in which to save stickers. giving `-` means stdout. defaults to `-`.')),
        (['-pdf'], dict(action='store_const', const=cairo.PDFSurface, default=None, dest='surface_cls', help='use pdf file format. by default, the format will be guessed from the file extension')),
        (['-ps'], dict(action='store_const', const=cairo.PSSurface, dest='surface_cls', help='use ps file format.')),
        (['items'], dict(nargs=argparse.REMAINDER, default='',
            help='items to generate stickers for. This can be one of '
                 '{section, book, library}'
            )),
    )
    
    SURFACE_EXT = {
        '.ps': cairo.PSSurface,
        '': cairo.PDFSurface
    }
    
    main_parser = argparse.ArgumentParser(prog='stickers')
    subparsers = main_parser.add_subparsers()
    
    parser = subparsers.add_parser('section')
    parser.set_defaults(cls=db.Section)
    parser.add_argument('-c', '--count', type=int, default=1)
    parser.add_argument('-r', '--range', nargs=2, action='append', default=[], dest='filters')
    parser.add_argument('-p', '--pattern', action='append', dest='filters')
    parser.add_argument('other_items', nargs=argparse.REMAINDER)
    
    parser = subparsers.add_parser('book')
    parser.set_defaults(cls=db.Book)
    parser.add_argument('-c', '--count', type=int, default=1)
    parser.add_argument('-r', '--range', nargs=2, action='append', default=[], dest='filters')
    parser.add_argument('-p', '--pattern', action='append', dest='filters')
    parser.add_argument('other_items', nargs=argparse.REMAINDER)
    
    parser = subparsers.add_parser('library')
    parser.set_defaults(cls=db.Library)
    parser.add_argument('-c', '--count', type=int, default=1)
    parser.add_argument('-p', '--pattern', action='append', default=[], dest='filters')
    parser.add_argument('other_items', nargs=argparse.REMAINDER)
    
    parser = subparsers.add_parser('blank')
    parser.set_defaults(cls=None)
    parser.set_defaults(filters=[])
    parser.add_argument('-c', '--count', type=int, default=1, dest='count')
    parser.add_argument('other_items', nargs=argparse.REMAINDER)
    
    def do_call(self, file=None, items=[], surface_cls=None, ctx=None,
            queries=None, config=None):
        if cairo is None:
            self._message = 'cairo library is missing. must install it.'
            raise RuntimeError(self.message)
        
        if file == '-':
            file = sys.stdout.buffer
        
        if config is None:
            config = {}
        for c in self.CONFIG:
            if c not in config:
                config[c] = type(self.CONFIG[c])(
                    self.util.get_config(c))
        
        cls_filters = {
            db.Section:
                {
                    list: lambda x: text('full_number BETWEEN :num_1 AND :num_2').bindparams(num_1=x[0], num_2=x[1]),
                    str: lambda x: text('section_name like :pattern').bindparams(pattern=x),
                },
            db.Book:
                {
                    list: lambda x: text('section_full_number || "-" || book.number BETWEEN :num_1 AND :num_2').bindparams(num_1=x[0], num_2=x[1]),
                    str: lambda x: db.Book.title.like(x),
                },
            db.Library:
                {
                    str: lambda x: db.Library.name.like(x),
                },
            }
        
        cls_queries = {
                db.Section: self.section_query(),
                db.Book: self.book_query(),
                db.Library: self.util.get_scoped_session().query(db.Library),
                None: self.util.get_scoped_session().query('NULL')
            }
        
        found_items = []
        while items:
            ns = vars(type(self).main_parser.parse_args(items))
            items = ns.pop('other_items')
            found_items.append(ns)
        
        if queries is None:
            queries = []
            for item in found_items:
                cls = item['cls']
                filters = item['filters']
                
                q = {}
                q['count'] = item['count']
                q['q'] = cls_queries[cls]
                if filters:
                    cond = filters.pop(0)
                    q['q'] = q['q'].filter(cls_filters[cls][type(cond)](cond))
                queries.append(q)
                
                for c in filters:
                    q = {}
                    q['count'] = item['count']
                    q['q'] = cls_queries[cls].filter(
                        cls_filters[cls][type(cond)](cond))
                    queries.append(q)
        
        self._finished = 0
        self._total = sum([c['count'] * c['q'].count() for c in queries])
        
        try:
            sur_finish = False
            if ctx is None:
                MM_IN_PT = 72 / 25.4
                ext = pathlib.Path(file).suffix
                if surface_cls is None:
                    if ext in type(self).SURFACE_EXT:
                        surface_cls = type(self).SURFACE_EXT[ext]
                    else:
                        surface_cls = type(self).SURFACE_EXT['']
                
                sur = surface_cls(file,
                    MM_IN_PT * self.util.get_config('stickers_page_width'),
                    MM_IN_PT * self.util.get_config('stickers_page_height'))
                ctx = cairo.Context(sur)
                sur_finish = True
            
            self.gen_stickers(ctx, *queries, config=config)
        except Exception as e:
            self._message = str(e)
            raise
        finally:
            if sur_finish:
                sur.finish()
        
    def gen_stickers(self, ctx, *queries, config=None):
        if config is None:
            config = {}
        for c in self.CONFIG:
            if c not in config:
                config[c] = type(self.CONFIG[c])(
                    self.util.get_config(c))
        
        MM_IN_PT = 72 / 25.4
        config['stickers_top_margin'] *= MM_IN_PT
        config['stickers_left_margin'] *= MM_IN_PT
        config['stickers_col_sep'] *= MM_IN_PT
        config['stickers_row_sep'] *= MM_IN_PT
        config['stickers_border_width'] *= MM_IN_PT
        config['stickers_page_width'] *= MM_IN_PT
        config['stickers_page_height'] *= MM_IN_PT
        
        w = (config['stickers_page_width'] - 2 * config['stickers_left_margin'] -
                (config['stickers_columns'] - 1) * 2 *
                config['stickers_col_sep']
            ) / config['stickers_columns']
        h = (config['stickers_page_height'] - 2 * config['stickers_top_margin'] -
                (config['stickers_rows'] - 1) * 2 * config['stickers_row_sep']
            ) / config['stickers_rows']
        
        ctx.set_line_width(config['stickers_border_width'])
        ctx.set_source_rgb(*config['stickers_border_color'])
        
        row = 0
        column = 0
        x = config['stickers_hoffset'] + config['stickers_left_margin']
        y = config['stickers_voffset'] + config['stickers_top_margin']
        for q in queries:
            for c in q['q']:
                for d in range(q['count']):
                    if c != (None,):
                        type_name = type(c).__name__.lower()
                        template_name = 'stickers_{}_template'.format(type_name)
                        self.gen_sticker(ctx, c, x=x, y=y, w=w, h=h,
                            font=config['stickers_font'],
                            text_color=config['stickers_text_color'],
                            template=config[template_name])
                    
                    self.next()
                    
                    if self._stop:
                        break
                    
                    x += w + 2 * config['stickers_col_sep']
                    column += 1
                    if column >= config['stickers_columns']:
                        column = 0
                        x = (config['stickers_hoffset'] +
                            config['stickers_left_margin'])
                        row += 1
                        y += h + 2 * config['stickers_row_sep']
                        if row >= config['stickers_rows']:
                            row = 0
                            y = (config['stickers_voffset'] +
                                config['stickers_top_margin'])
                            ctx.get_target().show_page()
                
                if self._stop:
                    break
        
        if row > 0:
            ctx.get_target().show_page()
        
        self._message = 'done'
    
    def gen_sticker(self, ctx, obj, x, y, w, h, font, text_color, template):
        ctx.save()
        
        ctx.rectangle(x, y, w, h)
        ctx.stroke_preserve()
        ctx.clip()
        
        lout = PangoCairo.create_layout(ctx)
        lout.set_font_description(Pango.FontDescription(font))
        lout.set_width(Pango.SCALE * w)
        lout.set_alignment(Pango.Alignment.CENTER)
        
        type_name = type(obj).__name__.lower()
        text = template.format(**{type_name: obj})
        lout.set_markup(text, -1)
        extent = lout.get_size()
        xtrans = x
        ytrans = y - extent.height / Pango.SCALE / 2 + h / 2
        
        ctx.translate(xtrans, ytrans)
        ctx.set_source_rgb(*text_color)
        PangoCairo.update_layout(ctx, lout)
        PangoCairo.show_layout(ctx, lout)
        ctx.translate(-xtrans, -ytrans)
        ctx.restore()
    
    def section_query(self):
        session = self.util.get_scoped_session()
        root_section = self.util.get_root_section(session)
        
        digits = self.util.get_section_digits()
        padding = digits * '0'
        full_number = func.substr(padding + db.Section.number.cast(String),
            -digits, digits)
        icteq = session.query(
                db.Section.id,
                full_number.label('full_number'),
            ).filter_by(
                parent=root_section
            ).cte()
        rcteq = session.query(
                db.Section.id,
                icteq.c.full_number.concat('-').concat(full_number),
            ).join(
                icteq, icteq.c.id == db.Section.parent_id
            )
        cteq = icteq.union(rcteq)
        q = session.query(db.Section).join(
            cteq, db.Section.id == cteq.c.id).order_by(cteq.c.full_number)
        return q
    
    def book_query(self):
        session = self.util.get_scoped_session()
        root_section = self.util.get_root_section(session)
        
        digits = self.util.get_section_digits()
        padding = digits * '0'
        full_number = func.substr(padding + db.Section.number.cast(String),
            -digits, digits)
        icteq = session.query(
                db.Section.id,
                full_number.label('section_full_number'),
            ).filter_by(
                parent=root_section
            ).cte()
        rcteq = session.query(
                db.Section.id,
                icteq.c.section_full_number.concat('-').concat(full_number),
            ).join(
                icteq, icteq.c.id == db.Section.parent_id
            )
        cteq = icteq.union(rcteq)
        
        digits = self.util.get_book_digits()
        padding = digits * '0'
        full_number = cteq.c.section_full_number.concat('-').concat(
            func.substr(padding + db.Book.number.cast(String),
                -digits, digits))
        
        q = session.query(db.Book).join(cteq).order_by(
            cteq.c.section_full_number, db.Book.number)
        return q

